// SPDX-License-Identifier: BSD-3-Clause

/**
 * JDK9 removed the Nameservice SPI as per bug 8134577. The interfaces in this package were
 * automatically extracted and are used to compile on newer JDKs.
 */
package sun.net.spi.nameservice;
