// SPDX-License-Identifier: BSD-3-Clause

/**
 * The interfaces in this package were automatically extracted and are used to compile on against
 * the Android SDK without installing it. The RoboElectric dependency does not work with JPMS.
 */
package android;
